// Copyright 2023 Google LLC
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

#![allow(unused)]

use core::fmt::{self, Write};
use core::ptr::{addr_of, addr_of_mut};

// ANCHOR: Flags
use bitflags::bitflags;

bitflags! {
    /// _Flags_ do registrador de _flags_ do UART.
    #[repr(transparent)]
    #[derive(Copy, Clone, Debug, Eq, PartialEq)]
    struct Flags: u16 {
        /// Limpar para enviar.
        const CTS = 1 << 0;
        /// Dados prontos para envio.
        const DSR = 1 << 1;
        /// Dados detectados pelo receptor.
        const DCD = 1 << 2;
        /// UART ocupado transmitindo dados.
        const BUSY = 1 << 3;
        /// O FIFO de recebimento está vazio.
        const RXFE = 1 << 4;
        /// O FIFO de transmissão está cheio.
        const TXFF = 1 << 5;
        /// O FIFO de recebimento está cheio.
        const RXFF = 1 << 6;
        /// O FIFO de transmissão está vazio.
        const TXFE = 1 << 7;
        /// Indicador de anel.
        const RI = 1 << 8;
    }
}
// ANCHOR_END: Flags

bitflags! {
    /// Flags do Registrador de Status de Recebimento / Registrador de Limpeza de Erro UART.
    #[repr(transparent)]
    #[derive(Copy, Clone, Debug, Eq, PartialEq)]
    struct ReceiveStatus: u16 {
        /// Erro de enquadramento.
        const FE = 1 << 0;
        /// Erro de paridade.
        const PE = 1 << 1;
        /// Erro de interrupção.
        const BE = 1 << 2;
        /// Erro de estouro.
        const OE = 1 << 3;
    }
}

// ANCHOR: Registers
#[repr(C, align(4))]
struct Registers {
    dr: u16,
    _reserved0: [u8; 2],
    rsr: ReceiveStatus,
    _reserved1: [u8; 19],
    fr: Flags,
    _reserved2: [u8; 6],
    ilpr: u8,
    _reserved3: [u8; 3],
    ibrd: u16,
    _reserved4: [u8; 2],
    fbrd: u8,
    _reserved5: [u8; 3],
    lcr_h: u8,
    _reserved6: [u8; 3],
    cr: u16,
    _reserved7: [u8; 3],
    ifls: u8,
    _reserved8: [u8; 3],
    imsc: u16,
    _reserved9: [u8; 2],
    ris: u16,
    _reserved10: [u8; 2],
    mis: u16,
    _reserved11: [u8; 2],
    icr: u16,
    _reserved12: [u8; 2],
    dmacr: u8,
    _reserved13: [u8; 3],
}
// ANCHOR_END: Registers

// ANCHOR: Uart
/// Driver para um UART PL011.
#[derive(Debug)]
pub struct Uart {
    registers: *mut Registers,
}

impl Uart {
    /// Constrói uma nova instância do driver UART para um dispositivo PL011 no endereço
    /// base fornecido.
    ///
    /// # Segurança
    ///
    /// O endereço base fornecido deve apontar para os registradores de controle MMIO de um
    /// dispositivo PL011, que deve ser mapeado no espaço de endereço do processo
    /// como memória de dispositivo e não ter nenhum outro _alias_.
    pub unsafe fn new(base_address: *mut u32) -> Self {
        Self { registers: base_address as *mut Registers }
    }

    /// Grava um único byte no UART.
    pub fn write_byte(&self, byte: u8) {
        // Aguarde até que haja espaço no buffer TX.
        while self.read_flag_register().contains(Flags::TXFF) {}

        // Seguro porque sabemos que self.registers aponta para o controle
        // registradores de um dispositivo PL011 que está mapeado adequadamente.
        unsafe {
            // Escreva no buffer TX.
            addr_of_mut!((*self.registers).dr).write_volatile(byte.into());
        }

        // Aguarde até que o UART não esteja mais ocupado.
        while self.read_flag_register().contains(Flags::BUSY) {}
    }

    /// Lê e retorna um byte pendente ou `None` se nada foi
    /// recebido.
    pub fn read_byte(&self) -> Option<u8> {
        if self.read_flag_register().contains(Flags::RXFE) {
            None
        } else {
            let data = unsafe { addr_of!((*self.registers).dr).read_volatile() };
            // TODO: Verifique as condições de erro nos bits 8-11.
            Some(data as u8)
        }
    }

    fn read_flag_register(&self) -> Flags {
        // Seguro porque sabemos que self.registers aponta para o controle
        // registradores de um dispositivo PL011 que está mapeado adequadamente.
        unsafe { addr_of!((*self.registers).fr).read_volatile() }
    }
}
// ANCHOR_END: Uart

impl Write for Uart {
    fn write_str(&mut self, s: &str) -> fmt::Result {
        for c in s.as_bytes() {
            self.write_byte(*c);
        }
        Ok(())
    }
}

// Seguro porque ele contém apenas um ponteiro para memória de dispositivo, que pode ser
// acessado de qualquer contexto.
unsafe impl Send for Uart {}
